from typing import Union, List
import inspect


def get_function_info(func):
    signature = inspect.signature(func)
    parameters = signature.parameters

    func_info = {
        "name": func.__name__,
        "description": func.__doc__,
        "parameters": {
            "type": "object",
            "properties": {}
        }
    }

    for param_name, param in parameters.items():
        param_info = {
            "type": str(param.annotation),
            "description": param.default.__doc__ if param.default else None
        }
        func_info["parameters"][param_name] = param_info

    return func_info


class Tool:
    name: str
    description: str = None
    params = []
    function: callable = None

    def __init__(self, name=None, description=None, params=None, function=None):
        self.name = name or self.name
        self.description = description or self.description
        self.params = params or self.params
        self.function = function or self.function
    
    @classmethod
    def parse(cls, tool):
        if isinstance(tool, Tool):
            return tool
        elif callable(tool):
            tool_info = get_function_info(tool)
            tool = Tool(**{'function': tool, **tool_info})
            return tool
    
    @property
    def info(self):
        func_info = {
            "name": self.name,
            "description": self.description,
            "params": self.params,
        }
        return func_info
    
    def exec(self, *args, **kwargs):
        return self.function(**kwargs)
    
    def __call__(self, *args, **kwargs):
        return self.exec(*args, **kwargs)


ToolList = Union[List[Tool], List[callable]]


import os
from loguru import logger
from rich import pretty
from rich.logging import RichHandler

from .world import World
from .api import API
from .admin import Admin
from .collection import ChromaVectorDB


class App:
    name: str
    path: str
    world: World

    def __init__(self, name, path, world=None, db=None):
        self.name = name
        self.path = path
        self.world = world or World(name, db=db)
    
    @classmethod
    def load(cls, path, db=None, env=None):
        db = db or ChromaVectorDB(path=path)

        config = {
            'name': 'local',
            'path': path,
            'db': db,
        }

        pretty.install()

        env = env or os.environ
        log_file_path = f"./log/{env.get('PX_ENV', 'development')}.log"  # Replace 'development' with your environment
        level = env.get('PX_LOG_LEVEL', 'INFO')  # Replace 'CRITICAL' with your log level

        # Configure Loguru
        logger.remove()
        logger.add(
            log_file_path,
            rotation="10 MB",  # Rotate log files that exceed 10 MB
            level=level,  # Minimum level to log
            format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | <level>{level: <8}</level> | <level>{message}</level>",
            backtrace=True  # Enable traceback
        )

        logger.info("Log file: " + log_file_path)
        logger.info("Log level: " + level)

        return cls(**config)

    @property
    def api(self):
        if self._api is None:
            self._api = API(self.world)
        return self._api

    @property
    def admin(self):
        if self._admin is None:
            self._admin = Admin(self.world)
        return self._admin
    
    def __repr__(self):
        return f'<App {self.name} path={self.path}>'




import os
import time
from typing import Callable, List

from .collection import Collection 
from .world import Session
from .application import App
from .auth import DefaultUser 


def prompt(instructions=None, input=None, output=None, prompt=None, context=None, template=None, llm=None, examples=None, num_examples=1, history=None, tools=None, dryrun=False, retries=3, debug=False, silent=False, **kwargs):
    kwargs = dict(
        instructions=instructions,
        input=input,
        output=output,
        prompt=prompt,
        llm=llm,
        context=context,
        examples=examples,
        num_examples=num_examples,
        history=history,
        tools=tools,
        template=template,
        debug=debug,
        dryrun=dryrun,
        silent=silent,
        retries=retries,
        store=store,
    )
    return DEFAULT_SESSION.prompt(**kwargs)


def chain(*steps, **kwargs):
    return DEFAULT_SESSION.chain(*steps, **kwargs)


def store(*items, collection=None, **kwargs) -> Collection:
    return DEFAULT_SESSION.store(*items, collection=collection, **kwargs)


def delete(*items, collection=None, **kwargs) -> None:
    return DEFAULT_SESSION.delete(*items, collection=collection, **kwargs)


def collections():
    return DEFAULT_SESSION.collections()


def delete_collection(collection, **kwargs) -> None:
    return DEFAULT_SESSION.delete_collection(collection, **kwargs)


def create_collection(collection, **kwargs) -> None:
    return DEFAULT_SESSION.create_collection(collection, **kwargs)


def query(*texts, field=None, where=None, collection=None, **kwargs) -> Collection:
    return DEFAULT_SESSION.query(
        *texts, field=field, where=where, collection=collection, **kwargs)


def chat(message, context=None, **kwargs):
    return DEFAULT_SESSION.chat(message, context, **kwargs)


def evaluate(*testcases, **kwargs) -> Collection:
    return DEFAULT_SESSION.evaluate(*testcases, **kwargs)


def history(query=None, **kwargs):
    if query is not None:
        return DEFAULT_SESSION.history(query, **kwargs)
    else:
        return DEFAULT_SESSION.history


def templates(ids=None, **kwargs) -> Collection:
    return DEFAULT_SESSION.world.templates(ids=ids, **kwargs)


def systems(ids=None, **kwargs) -> Collection:
    return DEFAULT_SESSION.world.systems(ids=ids, **kwargs)


def session() -> Session:
    return DEFAULT_SESSION


def run(world=None, **kwargs):
    if world is None:
        world = DEFAULT_WORLD
    return world(**kwargs)


DEFAULT_WORLD = None
def set_default_world(world):
    global DEFAULT_WORLD
    DEFAULT_WORLD = world


DEFAULT_SESSION = None
def set_default_session(session):
    global DEFAULT_SESSION
    DEFAULT_SESSION = session


Embedding = List[float]
EmbedFunction = Callable[[List[str]], List[Embedding]]


def find_project_root(path=None, config_filename='.px'):
    home_dir = os.path.expanduser("~")
    if path is None:
        path = os.getcwd()
    while path != home_dir:
        if os.path.exists(os.path.join(path, config_filename)) and os.path.isdir(path):
            return path
        path = os.path.dirname(path)
    if os.path.exists(os.path.join(home_dir, config_filename)):
        return home_dir
    return None


def load(path='local'):
    if path == 'local':
        path = find_project_root()

    if path is None:
        raise ValueError('could not find project')

    app = None
    if path.startswith('http'):
        print('loading remote app from', path)
        raise NotImplementedError
    else:
        from dotenv import load_dotenv
        load_dotenv(os.path.join(path, '.env'))
        app = App.load(path)

        # look for a config.py file and execute it
        config_file = os.path.join(path, 'config.py')
        if os.path.exists(config_file):
            import importlib.util
            spec = importlib.util.spec_from_file_location('config', config_file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            session = app.world.create_session(f'{app.name}_config_{int(time.time())}')
            module.setup(session)
    
    user = DefaultUser()
    s = app.world.create_session(user)
    set_default_session(s)
    return app


from enum import Enum
from .collection import Entity


class UserRole(Enum):
    admin = 'admin'


class User(Entity):
    role: UserRole
    username: str


class DefaultUser(User):
    role: UserRole = UserRole.admin
    username: str = 'root'


from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='pxx',
    version='0.0.1',
    description='An AI framework',
    packages=['promptx'],
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "px = promptx.cli:main",
        ],
    },
)


from typing import Any

from fastapi import FastAPI, HTTPException
from pydantic import BaseModel

from .world import World 
from .collection import Query
from .template import Template
from .auth import DefaultUser


class PromptInput(BaseModel):
    input: Any = None


class API:
    world: World

    def __init__(self, world):
        self.world = world
        self.fastapi_app = FastAPI()

        @self.fastapi_app.get("/inbox")
        async def get_inbox():
            user = DefaultUser()
            logs = self.world.logs(where={'assigned_to': user.id})
            logs = self.world.logs()

            if logs is not None:
                return {"list": logs.objects}
            else:
                return {"list": []}

        @self.fastapi_app.post("/prompt")
        async def run_prompt(details: Template):
            session = self.world.create_session()
            template = Template(**details.dict())
            response = session.prompt(**{**dict(template), 'input': {}})
            return {"response": response}
        
        @self.fastapi_app.get("/templates")
        async def get_templates():
            r = self.world.templates()
            if r is None:
                templates = []
            else:
                templates = r.objects
            
            return {"list": templates}

        @self.fastapi_app.get("/{collection}")
        async def get_index(collection: str, query: str = None):
            if collection == 'collections':
                c = self.world.collections
            else:
                c = self.world._collections[collection]
            if c is None:
                raise HTTPException(status_code=404, detail=f"{collection} collection not found")
            if query is not None:
                r = c(query).objects
            else:
                r = c.objects
            return {'list': r}

        @self.fastapi_app.get("/{collection}/{id}")
        async def get_entity(collection: str, id: str):
            if collection == 'collections':
                c = self.world.collections
            else:
                c = self.world._collections[collection]
            r = c(ids=[id]).first
            return {'details': r, 'collection': collection}

        @self.fastapi_app.get("/templates/{id}")
        async def get_template(id: str):
            template = self.world.templates(ids=[id]).first
            if template is None:
                raise HTTPException(status_code=404, detail="Template not found")
            return {'details': template}
        
        @self.fastapi_app.post("/templates")
        async def create_template(details: Template):
            t = self.world.create_template(details)
            return t.id

        @self.fastapi_app.post("/templates/{id}/run")
        async def run_template(id: str, input: PromptInput = None):
            session = self.world.create_session(DefaultUser, debug=True)
            response = session.prompt(template=id, input=input, to_json=True)
            return {"response": response}
        
        @self.fastapi_app.get("/logs")
        async def get_logs():
            if self.world.logs.empty:
                return {'list': []}
            else:
                return {'list': self.world.logs.objects}

        @self.fastapi_app.get("/conversations")
        async def get_collections():
            return {"list": []}

        @self.fastapi_app.get("/collections")
        async def get_collections():
            return {"list": [c for c in self.world.collections().objects]}

        @self.fastapi_app.get("/collections/{name}")
        async def get_collection(name: str):
            try:
                c = self.world.collections(ids=[name])
                cc = self.world._collections[name]
                if c is None:
                    return {"response": None}
                else:
                    return {"details": dict(c), "list": cc.objects}
            except KeyError as e:
                raise HTTPException(status_code=404, detail="Collection not found")

        @self.fastapi_app.get("/systems")
        async def get_systems():
            r = self.world.systems()
            if r is None:
                systems = []
            else:
                systems = r.objects
            
            return {"list": systems}
        
        @self.fastapi_app.get("/systems/{name}")
        async def get_system(name: str):
            system = self.world.systems(ids=[name]).first
            return {"details": system}
        
        @self.fastapi_app.post("/systems/run")
        async def run_systems():
            session = self.world.create_session()
            return {"response": self.world(session)}
        
        @self.fastapi_app.post("/query")
        async def query(query: Query):
            session = self.world.create_session()
            response = session.query(query.query, where=query.where, collection=query.collection)
            return {"response": response}